#include <stdio.h>
#include <stdlib.h>
//#include <sys/types.h>
//#include <sys/stat.h>
#include <fcntl.h>
#include <linux/videodev.h>

#include "v4l2uvc.h"

#define PIC_WIDTH			320
#define PIC_HEIGHT		240
#define PIC_SIZE			(PIC_WIDTH*PIC_HEIGHT)

unsigned char round_d(double d)
{
	if (d<=0) return 0;
	if (d>=255) return 255;
	return d;
}

void decode_yuv444(const unsigned char y, const unsigned char u, const unsigned char v, unsigned char* rgb)
{
	double Y = y - 16.0;
	double V = v - 128.0;
	double U = u - 128.0;

	rgb[0] = round_d(1.164*Y           + 2.018*U);
	rgb[1] = round_d(1.164*Y - 0.813*V - 0.391*U);
	rgb[2] = round_d(1.164*Y + 1.596*V);
} 

unsigned char round_fast(int val)
{
	if (val<=0) return 0;
	if (val>=65535) return 255;
	return val>>8;
}

void decode_yuv444_fast(const unsigned char y, const unsigned char u, const unsigned char v, unsigned char* rgb)
{
	int Y = (int)y - 16;
	int U = (int)u - 128;
	int V = (int)v - 128;
	rgb[0] = round_fast(298*Y 			  + 516*U + 128);
	rgb[1] = round_fast(298*Y - 208*V - 100*U + 128);
	rgb[2] = round_fast(298*Y + 409*V + 128);
} 

int main()
{
	int count;
	int i;
	int len;
	int total;
	int fdout;
	pid_t child;
	int status;
	unsigned char u, y1, v, y2;
	unsigned char rgb[6];
	unsigned char* p;
	struct timeval start;
	struct timeval end;
  int mtime, seconds, useconds;  

	FILE* f;
	int video_dev;
	struct vdIn* videoIn = (struct vdIn*)malloc(sizeof(struct vdIn));

	char* device = "/dev/video0" ; //domyslnie /dev/video

  /* open video device and prepare data structure */
  video_dev = init_videoIn(videoIn, device, 320, 240, 5, V4L2_PIX_FMT_MJPEG/*V4L2_PIX_FMT_YUYV*/, 1);
  if (video_dev < 0) {
    perror("init_VideoIn failed");
    return -1;
  }

  videoIn->tmpbuffer = (unsigned char *)malloc(videoIn->framesizeIn);
	if (!videoIn->tmpbuffer) {
		perror("Memory allocation error");
		return -2;	
	}

	printf("Video init...");
	fflush(stdout);
	if( uvcGrab(videoIn) < 0 ) {
	  perror("Error grabbing");
	  return -3;
	}
	printf("DONE\n");


	for (count=0;count<4;count++) {
		gettimeofday(&start, NULL);		
		if( uvcGrab(videoIn) < 0 ) {
	    perror("Error grabbing");
	    return -3;
	  }
		gettimeofday(&end, NULL);
    seconds  = end.tv_sec  - start.tv_sec;
    useconds = end.tv_usec - start.tv_usec;
    mtime = ((seconds) * 1000 + useconds/1000.0) + 0.5;
		printf("Grabbing time: %d ms\n", mtime);
		
		gettimeofday(&start, NULL);		
		f = fopen("test.pnm", "wb");
		fprintf(f, "P6\n");
		fprintf(f, "320 240\n");
		fprintf(f, "255\n");
		p = videoIn->tmpbuffer;
		for (i=0;i<PIC_SIZE/2;i++) {
				decode_yuv444(p[0], p[3], p[1], rgb);
				decode_yuv444(p[2], p[3], p[1], rgb+3);

				p += 4;
				fwrite(rgb, 1, 6, f);
			}
		fclose(f);
		gettimeofday(&end, NULL);
    seconds  = end.tv_sec  - start.tv_sec;
    useconds = end.tv_usec - start.tv_usec;
    mtime = ((seconds) * 1000 + useconds/1000.0) + 0.5;
		printf("Test time: %d ms\n", mtime);

		gettimeofday(&start, NULL);		
		f = fopen("test2.pnm", "wb");
		fprintf(f, "P6\n");
		fprintf(f, "320 240\n");
		fprintf(f, "255\n");
		p = videoIn->tmpbuffer;
		for (i=0;i<PIC_SIZE/2;i++) {
				decode_yuv444_fast(p[0], p[3], p[1], rgb);
				decode_yuv444_fast(p[2], p[3], p[1], rgb+3);

				p += 4;
				fwrite(rgb, 1, 6, f);
			}
		fclose(f);
		gettimeofday(&end, NULL);
    seconds  = end.tv_sec  - start.tv_sec;
    useconds = end.tv_usec - start.tv_usec;
    mtime = ((seconds) * 1000 + useconds/1000.0) + 0.5;
		printf("Test2 time: %d ms\n", mtime);

		gettimeofday(&start, NULL);		
		f = fopen("test3.pnm", "wb");
		fprintf(f, "P6\n");
		fprintf(f, "320 240\n");
		fprintf(f, "255\n");
		p = videoIn->tmpbuffer;
		for (i=0;i<PIC_SIZE/2;i++) {
				rgb[0] = rgb[1] = rgb[2] = p[0];
				rgb[3] = rgb[4] = rgb[5] = p[2];

				p += 4;
				fwrite(rgb, 1, 6, f);
			}
		fclose(f);
		gettimeofday(&end, NULL);
    seconds  = end.tv_sec  - start.tv_sec;
    useconds = end.tv_usec - start.tv_usec;
    mtime = ((seconds) * 1000 + useconds/1000.0) + 0.5;
		printf("Test3 time: %d ms\n", mtime);
	}

	return 0;
}

